/**
 * @file core/tree/cover_tree.hpp
 * @author Ryan Curtin
 *
 * Includes all the necessary files to use the CoverTree class.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_CORE_TREE_COVER_TREE_HPP
#define MLPACK_CORE_TREE_COVER_TREE_HPP

#include <mlpack/prereqs.hpp>
#include "cover_tree/first_point_is_root.hpp"
#include "cover_tree/cover_tree.hpp"
#include "cover_tree/single_tree_traverser.hpp"
#include "cover_tree/single_tree_traverser_impl.hpp"
#include "cover_tree/dual_tree_traverser.hpp"
#include "cover_tree/dual_tree_traverser_impl.hpp"
#include "cover_tree/traits.hpp"
#include "cover_tree/typedef.hpp"

#endif
