/**
 * @file methods/ann/regularizer/regularizer.hpp
 * @author Saksham Bansal
 *
 * This includes various regularizers to construct a model.
 *
 * mlpack is free software; you may redistribute it and/or modify it under the
 * terms of the 3-clause BSD license.  You should have received a copy of the
 * 3-clause BSD license along with mlpack.  If not, see
 * http://www.opensource.org/licenses/BSD-3-Clause for more information.
 */
#ifndef MLPACK_METHODS_ANN_REGULARIZER_HPP
#define MLPACK_METHODS_ANN_REGULARIZER_HPP

#include "lregularizer.hpp"
#include "orthogonal_regularizer.hpp"
#include "no_regularizer.hpp"

#endif
